from __future__ import division

from numpy import log
from numpy import exp
from numpy_sugar.special import (normal_cdf, normal_icdf)


class Link(object):
    def __init__(self):
        super(Link, self).__init__()

    def value(self, x):
        raise NotImplementedError

    def inv(self, x):
        raise NotImplementedError


class LogitLink(Link):
    def __init__(self):
        super(LogitLink, self).__init__()

    def value(self, x):
        return log(x / (1 - x))

    def inv(self, x):
        return 1 / (1 + exp(-x))


class ProbitLink(Link):
    def __init__(self):
        super(ProbitLink, self).__init__()

    def value(self, x):
        return normal_icdf(x)

    def inv(self, x):
        return normal_cdf(x)


class LogLink(Link):
    def __init__(self):
        super(LogLink, self).__init__()

    def value(self, x):
        return log(x)

    def inv(self, x):
        return exp(x)



from __future__ import division

from scipy.stats import multivariate_normal
from numpy import (dot, log, var, zeros, sqrt, ascontiguousarray)

from numpy_sugar.linalg import (sum2diag, solve, economic_svd, ddot)


class FastLMMCore(object):
    def __init__(self, y, M, Q0, Q1, S0):
        if var(y) < 1e-8:
            raise ValueError("The phenotype variance is too low: %e." % var(y))

        self._n = y.shape[0]
        self._p = self._n - S0.shape[0]
        self._S0 = S0
        self._diag0 = S0 * 0.5 + 0.5
        self._diag1 = 0.5
        self._Q0 = Q0
        self._Q1 = Q1
        self._tM = None
        self.__tbeta = None
        self._covariate_setup(M)
        self._M = M

        d = M.shape[1]
        self._scale = 1.0
        self._delta = 0.5
        self._lml = 0.0

        self._a1 = 0.0
        self._b1 = zeros(d)
        self._c1 = zeros((d, d))

        self._a0 = 0.0
        self._b0 = zeros(d)
        self._c0 = zeros((d, d))

        self._yTQ0 = dot(y.T, Q0)
        self._yTQ0_2x = self._yTQ0**2

        self._yTQ1 = dot(y.T, Q1)
        self._yTQ1_2x = self._yTQ1**2

        self._oneTQ0 = self._tM.T.dot(Q0)
        self._oneTQ1 = self._tM.T.dot(Q1)

        self._valid_update = 0
        self.__Q0tymD0 = None
        self.__Q1tymD1 = None

    def _covariate_setup(self, M):
        SVD = economic_svd(M)
        self._svd_U = SVD[0]
        self._svd_S12 = sqrt(SVD[1])
        self._svd_V = SVD[2]
        self._tM = ddot(self._svd_U, self._svd_S12, left=False)

    def copy(self):
        o = FastLMMCore.__new__(FastLMMCore)
        o._n = self._n
        o._p = self._p
        o._S0 = self._S0
        o._diag0 = self._diag0.copy()
        o._diag1 = self._diag1
        o._Q0 = self._Q0
        o._Q1 = self._Q1
        o._M = self._M

        o.__tbeta = self.__tbeta.copy()
        o._scale = self._scale
        o._delta = self._delta
        o._lml = self._lml

        o._a1 = self._a1
        o._b1 = self._b1.copy()
        o._c1 = self._c1.copy()

        o._a0 = self._a0
        o._b0 = self._b0.copy()
        o._c0 = self._c0.copy()

        o._yTQ0 = self._yTQ0
        o._yTQ0_2x = self._yTQ0_2x

        o._yTQ1 = self._yTQ1
        o._yTQ1_2x = self._yTQ1_2x

        o._oneTQ0 = self._oneTQ0
        o._oneTQ1 = self._oneTQ1

        o._valid_update = self._valid_update
        from copy import copy
        o.__Q0tymD0 = copy(self.__Q0tymD0)
        o.__Q1tymD1 = copy(self.__Q1tymD1)

        return o

    @property
    def M(self):
        return self._M

    @M.setter
    def M(self, v):
        self._M = v
        self._covariate_setup(v)
        d = self._tM.shape[1]
        self.__tbeta = zeros(d)

        self._b1 = zeros(d)
        self._c1 = zeros((d, d))

        self._b0 = zeros(d)
        self._c0 = zeros((d, d))

        self._oneTQ0 = self._tM.T.dot(self._Q0)
        self._oneTQ1 = self._tM.T.dot(self._Q1)

        self._valid_update = 0
        self.__Q0tymD0 = None
        self.__Q1tymD1 = None

    @property
    def m(self):
        r"""Returns :math:`\mathbf m = \mathrm M \boldsymbol\beta`."""
        return dot(self._tM, self._tbeta)

    def _Q0tymD0(self):
        if self.__Q0tymD0 is None:
            Q0tym = self._yTQ0 - self.__tbeta.dot(self._oneTQ0)
            self.__Q0tymD0 = Q0tym / self._diag0
        return self.__Q0tymD0

    def _Q1tymD1(self):
        if self.__Q1tymD1 is None:
            Q1tym = self._yTQ1 - self.__tbeta.dot(self._oneTQ1)
            self.__Q1tymD1 = Q1tym / self._diag1
        return self.__Q1tymD1

    @property
    def _tbeta(self):
        return self.__tbeta

    @_tbeta.setter
    def _tbeta(self, value):
        if self.__tbeta is None:
            self.__tbeta = ascontiguousarray(value, float).copy()
        else:
            self.__tbeta[:] = value

    @property
    def beta(self):
        r"""Returns :math:`\boldsymbol\beta`."""
        return solve(self._svd_V.T, self._tbeta / self._svd_S12)

    @beta.setter
    def beta(self, value):
        self._tbeta = self._svd_S12 * dot(self._svd_V.T, value)

    @property
    def scale(self):
        return self._scale

    @property
    def delta(self):
        return self._delta

    @delta.setter
    def delta(self, delta):
        self._valid_update = 0
        self.__Q0tymD0 = None
        self.__Q1tymD1 = None
        self._delta = delta

    def _update_joints(self):
        yTQ0_2x = self._yTQ0_2x
        yTQ1_2x = self._yTQ1_2x

        self._a1 = yTQ1_2x.sum() / self._diag1
        self._b1[:] = (self._yTQ1 / self._diag1).dot(self._oneTQ1.T)
        self._c1[:] = (self._oneTQ1 / self._diag1).dot(self._oneTQ1.T)

        self._a0 = (yTQ0_2x / self._diag0).sum()
        self._b0[:] = (self._yTQ0 / self._diag0).dot(self._oneTQ0.T)
        self._c0[:] = (self._oneTQ0 / self._diag0).dot(self._oneTQ0.T)

    def _update_fixed_effects(self):
        nominator = self._b1 - self._b0
        denominator = self._c1 - self._c0
        self._tbeta = solve(denominator, nominator)

    def _update_scale(self):
        b = self.__tbeta
        p0 = self._a1 - 2 * self._b1.dot(b) + b.dot(self._c1.dot(b))
        p1 = self._a0 - 2 * self._b0.dot(b) + b.dot(self._c0).dot(b)
        self._scale = (p0 + p1) / self._n

    def _update_diags(self):
        self._diag0[:] = self._S0
        self._diag0 *= (1 - self._delta)
        self._diag0 += self._delta
        self._diag1 = self._delta

    def _update(self):
        if self._valid_update:
            return

        self._update_diags()
        self._update_joints()
        self._update_fixed_effects()
        self._update_scale()

        self._valid_update = 1

    def lml(self):
        if self._valid_update:
            return self._lml

        self._update()

        n = self._n
        p = self._p
        LOG2PI = 1.837877066409345339081937709124758839607238769531250
        self._lml = -n * LOG2PI - n - n * log(self._scale)
        self._lml += -sum(log(self._diag0)) - p * log(self._diag1)
        self._lml /= 2
        return self._lml

    def predict(self, covariates, Cp, Cpp):
        delta = self.delta

        diag0 = self._diag0
        diag1 = self._diag1

        CpQ0 = Cp.dot(self._Q0)
        CpQ1 = Cp.dot(self._Q1)

        m = covariates.dot(self.beta)
        mean = m + (1 - delta) * CpQ0.dot(self._Q0tymD0())
        mean += (1 - delta) * CpQ1.dot(self._Q1tymD1())

        cov = sum2diag(Cpp * (1 - self.delta), self.delta)
        cov -= (1 - delta)**2 * CpQ0.dot((CpQ0 / diag0).T)
        cov -= (1 - delta)**2 * CpQ1.dot((CpQ1 / diag1).T)
        cov *= self.scale

        return FastLMMPredictor(mean, cov)


class FastLMMPredictor(object):
    def __init__(self, mean, cov):
        self._mean = mean
        self._cov = cov
        self._mvn = multivariate_normal(mean, cov)

    @property
    def mean(self):
        return self._mean

    @property
    def cov(self):
        return self._cov

    def pdf(self, y):
        return self._mvn.pdf(y)

    def logpdf(self, y):
        return self._mvn.logpdf(y)



import numpy as np
import numpy.testing as npt

from limix_inference.cov import LinearCov
from optimix import check_grad


def test_value():
    random = np.random.RandomState(0)
    cov = LinearCov()

    x0 = random.randn(10)
    x1 = random.randn(10)

    npt.assert_almost_equal(cov.value(x0, x1), np.dot(x0, x1))
    cov.scale = 2.
    npt.assert_almost_equal(cov.value(x0, x1), 2 * np.dot(x0, x1))


def test_gradient():
    random = np.random.RandomState(0)
    cov = LinearCov()
    cov.scale = 2.

    x0 = random.randn(10)
    x1 = random.randn(10)

    def func(x):
        cov.scale = np.exp(x[0])
        return cov.value(x0, x1)

    def grad(x):
        cov.scale = np.exp(x[0])
        return [cov.derivative_logscale(x0, x1)]

    npt.assert_almost_equal(check_grad(func, grad, [2.0]), 0, decimal=6)



from ._ffi import lib, ffi

from numpy import ascontiguousarray
from numpy import ndarray

def ptr(a):
    return ffi.cast("double *", a.ctypes.data)

class LikNormMachine(object):
    def __init__(self, npoints=500):
        self._machine = lib.create_machine(npoints)

    def finish(self):
        lib.destroy_machine(self._machine)

    def moments(self, likname, y, eta, tau, log_zeroth, mean, variance):
        lik = getattr(lib, likname.upper())

        tau = ascontiguousarray(tau, float)
        eta = ascontiguousarray(eta, float)
        log_zeroth = ascontiguousarray(log_zeroth, float)
        mean = ascontiguousarray(mean, float)
        variance = ascontiguousarray(variance, float)

        size = len(log_zeroth)

        # tau = ptr(tau)
        # eta = ptr(eta)
        # log_zeroth = ptr(log_zeroth)
        # mean = ptr(mean)
        # variance = ptr(variance)

        # y = [ptr(yi) for yi in y]

        args = y + (tau, eta, log_zeroth, mean, variance)

        if likname.lower() == 'binomial':
            lib.apply2d(self._machine, lik, size, *(ptr(arg) for arg in args))
        else:
            lib.apply1d(self._machine, lik, size, *(ptr(arg) for arg in args))

        import numpy as np
        if not np.all(np.isfinite(log_zeroth)):
            import pdb; pdb.set_trace()
            pass



from __future__ import division

from numpy import sum
from numpy import dot
from numpy import newaxis
from numpy import nan_to_num
from numpy_sugar.linalg import cho_solve
from numpy import errstate
from numpy_sugar.linalg import ddot


class FixedEP(object):
    def __init__(self, lml_const, A, C, L, Q, QBiQtCteta, teta, beta_nom):
        self._lml_const = lml_const
        self._QBiQtCteta = QBiQtCteta
        self._teta = teta
        self._A = A
        self._C = C
        self._L = L
        self._Q = Q
        self._beta_nom = beta_nom

    def compute(self, covariates, X):
        A = self._A
        L = self._L
        Q = self._Q
        C = self._C

        AMs0 = ddot(A, covariates, left=True)
        dens0 = AMs0 - ddot(A, dot(Q, cho_solve(L, dot(Q.T, AMs0))), left=True)
        noms0 = dot(self._beta_nom, covariates)

        AMs1 = ddot(A, X, left=True)
        dens1 = AMs1 - ddot(A, dot(Q, cho_solve(L, dot(Q.T, AMs1))), left=True)
        noms1 = dot(self._beta_nom, X)

        row00 = sum(covariates * dens0, 0)
        row01 = sum(covariates * dens1, 0)
        row11 = sum(X * dens1, 0)

        betas0 = noms0 * row11
        betas0 -= noms1 * row01
        betas1 = -noms0 * row01
        betas1 += noms1 * row00

        denom = row00 * row11
        denom -= row01**2

        with errstate(divide='ignore', invalid='ignore'):
            betas0 /= denom
            betas1 /= denom

        betas0 = nan_to_num(betas0)
        betas1 = nan_to_num(betas1)

        ms = dot(covariates, betas0[newaxis, :])
        ms += X * betas1
        QBiQtCteta = self._QBiQtCteta
        teta = self._teta

        Am = ddot(A, ms, left=True)
        w4 = dot(C * teta, ms)
        w4 -= dot(QBiQtCteta, Am)
        QBiQtAm = dot(Q, cho_solve(L, dot(Q.T, Am)))
        w5 = -sum(Am * ms, 0)
        w5 += sum(Am * QBiQtAm, 0)
        w5 /= 2
        lmls = self._lml_const + w4 + w5

        return (lmls, betas1)



from __future__ import division

from numpy_sugar.special import logbinom

from scipy.special import gammaln
import scipy.stats as st

from numpy import ascontiguousarray
from numpy import log1p
from numpy import exp
from numpy import log


class ProdLik(object):
    def __init__(self, likelihoods):
        super(ProdLik, self).__init__()
        self._likelihoods = likelihoods

    def pdf(self, x):
        return exp(self.logpdf(x))

    def logpdf(self, x):
        theta = self.theta(x)
        return (self.y * theta - self.b(theta)) / self.a() + self.c()

    @property
    def y(self):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].y for i in range(n)])

    @property
    def ytuple(self):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].ytuple for i in range(n)])

    def mean(self, x):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].mean(x[i]) for i in range(n)])

    def theta(self, x):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].teta(x[i]) for i in range(n)])

    @property
    def phi(self):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].phi for i in range(n)])

    def a(self):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].a() for i in range(n)])

    def b(self, theta):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].b(theta[i]) for i in range(n)])

    def c(self):
        liks = self._likelihoods
        n = len(liks)
        return _aca([liks[i].c() for i in range(n)])

    def sample(self, x):
        raise NotImplementedError


class DeltaProdLik(ProdLik):
    def __init__(self, link=None):
        super(DeltaProdLik, self).__init__(None)
        self._link = link
        self._outcome = None
        self.name = 'Delta'

    @property
    def outcome(self):
        return self._outcome

    @outcome.setter
    def outcome(self, v):
        self._outcome = _aca(v)

    @property
    def y(self):
        return self._outcome

    @property
    def ytuple(self):
        return (self._outcome,)

    def mean(self, x):
        return x

    def theta(self, x):
        return 0

    @property
    def phi(self):
        return 1

    def a(self):
        return 1

    def b(self, theta):
        return theta

    def c(self):
        return 0

    def sample(self, x, random_state=None):
        return self.mean(x)

    @property
    def sample_size(self):
        return len(self.outcome)


class BernoulliProdLik(ProdLik):
    def __init__(self, link):
        super(BernoulliProdLik, self).__init__(None)
        self._link = link
        self._outcome = None
        self.name = 'Bernoulli'

    @property
    def outcome(self):
        return self._outcome

    @outcome.setter
    def outcome(self, v):
        self._outcome = _aca(v)

    @property
    def y(self):
        return self._outcome

    @property
    def ytuple(self):
        return (self._outcome,)

    def mean(self, x):
        return self._link.inv(x)

    def theta(self, x):
        m = self.mean(x)
        return log(m / (1 - m))

    @property
    def phi(self):
        return 1

    def a(self):
        return 1 / self.phi

    def b(self, theta):
        return theta + log1p(exp(-theta))

    def c(self):
        return logbinom(self.phi, self.y * self.phi)

    def sample(self, x, random_state=None):
        p = self.mean(x)
        return st.bernoulli(p).rvs(random_state=random_state)

    def to_normal(self):
        y = self.outcome / self.outcome.std()
        y -= y.mean()
        return y

    @property
    def sample_size(self):
        return len(self.outcome)


class BinomialProdLik(ProdLik):
    def __init__(self, ntrials, link):
        super(BinomialProdLik, self).__init__(None)
        self._link = link
        self._nsuccesses = None
        self._ntrials = _aca(ntrials)
        self.name = 'Binomial'

    @property
    def ntrials(self):
        return self._ntrials

    @property
    def nsuccesses(self):
        return self._nsuccesses

    @nsuccesses.setter
    def nsuccesses(self, v):
        self._nsuccesses = _aca(v)

    @property
    def y(self):
        return self._nsuccesses / self._ntrials

    @property
    def ytuple(self):
        return (self._nsuccesses, self._ntrials)

    def mean(self, x):
        return self._link.inv(x)

    def theta(self, x):
        m = self.mean(x)
        return log(m / (1 - m))

    @property
    def phi(self):
        return self._ntrials

    def a(self):
        return 1 / self.phi

    def b(self, theta):
        return theta + log1p(exp(-theta))

    def c(self):
        return logbinom(self.phi, self.y * self.phi)

    def sample(self, x, random_state=None):
        p = self.mean(x)
        nt = ascontiguousarray(self._ntrials, dtype=int)
        return st.binom(nt, p).rvs(random_state=random_state)

    def to_normal(self):
        y = self.nsuccesses / self.ntrials
        y = y / y.std()
        y -= y.mean()
        return y

    @property
    def sample_size(self):
        return len(self.nsuccesses)


class PoissonProdLik(ProdLik):
    def __init__(self, link):
        super(PoissonProdLik, self).__init__(None)
        self._link = link
        self._noccurrences = None
        self.name = 'Poisson'

    @property
    def noccurrences(self):
        return self._noccurrences

    @noccurrences.setter
    def noccurrences(self, v):
        self._noccurrences = _aca(v)

    @property
    def y(self):
        return self._noccurrences

    @property
    def ytuple(self):
        return (self._noccurrences,)

    def mean(self, x):
        return self._link.inv(x)

    def theta(self, x):
        m = self.mean(x)
        return log(m)

    @property
    def phi(self):
        return 1

    def a(self):
        return self.phi

    def b(self, theta):
        return exp(theta)

    def c(self):
        return gammaln(self._noccurrences + 1)

    def sample(self, x, random_state=None):
        lam = self.mean(x)
        return st.poisson(mu=lam).rvs(random_state=random_state)

    def to_normal(self):
        y = self.noccurrences
        return (y - y.mean()) / y.std()

    @property
    def sample_size(self):
        return len(self.noccurrences)


def _aca(x):
    return ascontiguousarray(x, dtype=float)



import numpy as np
import numpy.testing as npt

from limix_inference.cov import SumCov
from limix_inference.cov import LinearCov
from optimix import check_grad


def test_value():
    random = np.random.RandomState(0)
    cov_left = LinearCov()
    cov_right = LinearCov()

    X0 = random.randn(4, 20)
    X1 = random.randn(4, 15)

    cov_left.set_data((X0, X0))
    cov_right.set_data((X1, X1))

    cov = SumCov([cov_left, cov_right])
    K = cov.feed().value()
    npt.assert_almost_equal(K[0, 0], 37.95568923)
    npt.assert_almost_equal(K[3, 1], 4.53034295)


def test_gradient():
    random = np.random.RandomState(0)
    cov_left = LinearCov()
    cov_right = LinearCov()

    X0 = random.randn(4, 20)
    X1 = random.randn(4, 15)

    cov_left.set_data((X0, X0))
    cov_right.set_data((X1, X1))

    cov = SumCov([cov_left, cov_right])

    def func(x):
        cov_left.scale = np.exp(x[0])
        cov_right.scale = np.exp(x[1])
        return cov.feed().value()

    def grad(x):
        cov_left.scale = np.exp(x[0])
        cov_right.scale = np.exp(x[1])
        return cov.feed().gradient()

    npt.assert_almost_equal(check_grad(func, grad, [2.0, 1.5]), 0, decimal=6)
